import hmac
import hashlib
import base64

from dydx3.constants import COLLATERAL_ASSET
from dydx3.constants import COLLATERAL_TOKEN_DECIMALS
from dydx3.constants import FACT_REGISTRY_CONTRACT
from dydx3.constants import TIME_IN_FORCE_GTT
from dydx3.constants import TOKEN_CONTRACTS
from dydx3.helpers.db import get_account_id
from dydx3.helpers.request_helpers import epoch_seconds_to_iso
from dydx3.helpers.request_helpers import generate_now_iso
from dydx3.helpers.request_helpers import generate_query_path
from dydx3.helpers.request_helpers import random_client_id
from dydx3.helpers.request_helpers import iso_to_epoch_seconds
from dydx3.helpers.request_helpers import json_stringify
from dydx3.helpers.request_helpers import remove_nones
from dydx3.helpers.requests import request
from dydx3.starkex.helpers import get_transfer_erc20_fact
from dydx3.starkex.helpers import nonce_from_client_id
from dydx3.starkex.order import SignableOrder
from dydx3.starkex.withdrawal import SignableWithdrawal
from dydx3.starkex.conditional_transfer import SignableConditionalTransfer
from dydx3.starkex.transfer import SignableTransfer


class Private(object):

    def __init__(
        self,
        host,
        network_id,
        stark_private_key,
        default_address,
        api_key_credentials,
    ):
        self.host = host
        self.network_id = network_id
        self.stark_private_key = stark_private_key
        self.default_address = default_address
        self.api_key_credentials = api_key_credentials

    # ============ Request Helpers ============

    def _private_request(
        self,
        method,
        endpoint,
        data={},
    ):
        now_iso_string = generate_now_iso()
        request_path = '/'.join(['/v3', endpoint])
        signature = self.sign(
            request_path=request_path,
            method=method.upper(),
            iso_timestamp=now_iso_string,
            data=remove_nones(data),
        )
        headers = {
            'DYDX-SIGNATURE': signature,
            'DYDX-API-KEY': self.api_key_credentials['key'],
            'DYDX-TIMESTAMP': now_iso_string,
            'DYDX-PASSPHRASE': self.api_key_credentials['passphrase'],
        }
        return request(
            self.host + request_path,
            method,
            headers,
            data,
        )

    def _get(self, endpoint, params):
        return self._private_request(
            'get',
            generate_query_path(endpoint, params),
        )

    def _post(self, endpoint, data):
        return self._private_request(
            'post',
            endpoint,
            data
        )

    def _put(self, endpoint, data):
        return self._private_request(
            'put',
            endpoint,
            data
        )

    def _delete(self, endpoint, params):
        return self._private_request(
            'delete',
            generate_query_path(endpoint, params),
        )

    # ============ Requests ============

    def get_api_keys(
        self,
    ):
        '''
        Get API keys.

        :returns: Object containing an array of apiKeys

        :raises: DydxAPIError
        '''
        return self._get(
            'api-keys',
            {},
        )

    def get_registration(self):
        '''
        Get signature for registration

        :returns: str

        :raises: DydxAPIError
        '''

        return self._get('registration', {})

    def get_user(self):
        '''
        Get user information

        :returns: User

        :raises: DydxAPIError
        '''
        return self._get('users', {})

    def update_user(
        self,
        user_data={},
        email=None,
        username=None,
        is_sharing_username=None,
        is_sharing_address=None,
        country=None,
        language_code=None,
    ):
        '''
        Update user information

        :param user_data: optional
        :type user_data: dict

        :param email: optional
        :type email: str

        :param username: optional
        :type username: str

        :param is_sharing_username: optional
        :type is_sharing_username: str

        :param is_sharing_address: optional
        :type is_sharing_address: str

        :param country optional
        :type country: str (ISO 3166-1 Alpha-2)

        :param language_code optional
        :type language_code: str (ISO 639-1, including 'zh-CN')

        :returns: User

        :raises: DydxAPIError
        '''
        return self._put(
            'users',
            {
                'email': email,
                'username': username,
                'isSharingUsername': is_sharing_username,
                'isSharingAddress': is_sharing_address,
                'userData': json_stringify(user_data),
                'country': country,
            },
        )

    def create_account(
        self,
        stark_public_key,
        stark_public_key_y_coordinate,
    ):
        '''
        Make an account

        :param stark_public_key: required
        :type stark_public_key: str

        :param stark_public_key_y_coordinate: required
        :type stark_public_key_y_coordinate: str

        :returns: Account

        :raises: DydxAPIError
        '''
        return self._post(
            'accounts',
            {
                'starkKey': stark_public_key,
                'starkKeyYCoordinate': stark_public_key_y_coordinate,
            }
        )

    def get_account(
        self,
        ethereum_address=None,
    ):
        '''
        Get an account

        :param ethereum_address: optional
        :type ethereum_address: str

        :returns: Account

        :raises: DydxAPIError
        '''
        address = ethereum_address or self.default_address
        if address is None:
            raise ValueError('ethereum_address was not set')
        return self._get(
            '/'.join(['accounts', get_account_id(address)]),
            {},
        )

    def get_accounts(
        self,
    ):
        '''
        Get accounts

        :returns: Array of accounts for a user

        :raises: DydxAPIError
        '''
        return self._get(
            'accounts',
            {},
        )

    def get_positions(
        self,
        market=None,
        status=None,
        limit=None,
        created_before_or_at=None,
    ):
        '''
        Get positions

        :param market: optional
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param status: optional
        :type status: str in list [
            "OPEN",
            "CLOSED",
            "LIQUIDATED",
        ]

        :param limit: optional
        :type limit: str

        :param created_before_or_at: optional
        :type created_before_or_at: ISO str

        :returns: Array of positions

        :raises: DydxAPIError
        '''
        return self._get(
            'positions',
            {
                'market': market,
                'limit': limit,
                'status': status,
                'createdBeforeOrAt': created_before_or_at,
            },
        )

    def get_orders(
        self,
        market=None,
        status=None,
        side=None,
        order_type=None,
        limit=None,
        created_before_or_at=None,
        returnLatestOrders=None,
    ):
        '''
        Get orders

        :param market: optional
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param status: optional
        :type status: str in list [
            "PENDING",
            "OPEN",
            "FILLED",
            "CANCELED",
            "UNTRIGGERED",
        ]

        :param side: optional
        :type side: str in list [
            "BUY",
            "SELL",
        ]

        :param order_type: optional
        :type order_type: str in list [
            "LIMIT",
            "STOP",
            "TRAILING_STOP",
            "TAKE_PROFIT",
        ]

        :param limit: optional
        :type limit: str

        :param created_before_or_at: optional
        :type created_before_or_at: ISO str

        :param returnLatestOrders: optional
        :type returnLatestOrders: boolean

        :returns: Array of Orders

        :raises: DydxAPIError
        '''
        return self._get(
            'orders',
            {
                'market': market,
                'status': status,
                'side': side,
                'type': order_type,
                'limit': limit,
                'createdBeforeOrAt': created_before_or_at,
                'returnLatestOrders': returnLatestOrders,
            },
        )

    def get_active_orders(
        self,
        market,
        side=None,
        id=None,
    ):
        '''
        Get ActiveOrders

        :param market: required
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param side: optional (required if id is passed in)
        :type side: str in list [
            "BUY",
            "SELL",
        ]

        param id: optional
        :type id: str

        :returns: Array of ActiveOrders

        :raises: DydxAPIError
        '''
        return self._get(
            'active-orders',
            {
                'market': market,
                'side': side,
                'id': id,
            },
        )

    def get_order_by_id(
        self,
        order_id,
    ):
        '''
        Get order by its id

        :param order_id: required
        :type order_id: str

        :returns: Order

        :raises: DydxAPIError
        '''
        return self._get(
            '/'.join(['orders', order_id]),
            {},
        )

    def get_order_by_client_id(
        self,
        client_id,
    ):
        '''
        Get order by its client_id

        :param client_id: required
        :type client_id: str

        :returns: Order

        :raises: DydxAPIError
        '''
        return self._get(
            '/'.join(['orders/client', client_id]),
            {},
        )

    def create_order(
        self,
        position_id,
        market,
        side,
        order_type,
        post_only,
        size,
        price,
        limit_fee,
        time_in_force=None,
        cancel_id=None,
        trigger_price=None,
        trailing_percent=None,
        client_id=None,
        expiration=None,
        expiration_epoch_seconds=None,
        signature=None,
    ):
        '''
        Post an order

        :param position_id: required
        :type position_id: str or int

        :param market: required
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param side: required
        :type side: str in list[
            "BUY",
            "SELL",
        ],

        :param order_type: required
        :type order_type: str in list [
            "LIMIT",
            "STOP",
            "TRAILING_STOP",
            "TAKE_PROFIT",
        ]

        :param post_only: required
        :type post_only: bool

        :param size: required
        :type size: str

        :param price: required
        :type price: str

        :param limit_fee: required
        :type limit_fee: str

        :param time_in_force: optional
        :type time_in_force: str in list [
            "GTT",
            "FOK",
            "IOC",
        ]

        :param cancel_id: optional
        :type cancel_id: str

        :param trigger_price: optional
        :type trigger_price: Decimal

        :param trailing_percent: optional
        :type trailing_percent: Decimal

        :param client_id: optional
        :type client_id: str

        :param expiration: optional
        :type expiration: ISO str

        :param expiration_epoch_seconds: optional
        :type expiration_epoch_seconds: int

        :param signature: optional
        type signature: str

        :returns: Order

        :raises: DydxAPIError
        '''
        client_id = client_id or random_client_id()
        if bool(expiration) == bool(expiration_epoch_seconds):
            raise ValueError(
                'Exactly one of expiration and expiration_epoch_seconds must '
                'be specified',
            )
        expiration = expiration or epoch_seconds_to_iso(
            expiration_epoch_seconds,
        )
        expiration_epoch_seconds = (
            expiration_epoch_seconds or iso_to_epoch_seconds(expiration)
        )

        order_signature = signature
        if not order_signature:
            if not self.stark_private_key:
                raise Exception(
                    'No signature provided and client was not ' +
                    'initialized with stark_private_key'
                )
            order_to_sign = SignableOrder(
                network_id=self.network_id,
                position_id=position_id,
                client_id=client_id,
                market=market,
                side=side,
                human_size=size,
                human_price=price,
                limit_fee=limit_fee,
                expiration_epoch_seconds=expiration_epoch_seconds,
            )
            order_signature = order_to_sign.sign(self.stark_private_key)

        order = {
            'market': market,
            'side': side,
            'type': order_type,
            'timeInForce': time_in_force or TIME_IN_FORCE_GTT,
            'size': size,
            'price': price,
            'limitFee': limit_fee,
            'expiration': expiration,
            'cancelId': cancel_id,
            'triggerPrice': trigger_price,
            'trailingPercent': trailing_percent,
            'postOnly': post_only,
            'clientId': client_id,
            'signature': order_signature,
        }

        return self._post(
            'orders',
            order,
        )

    def cancel_order(
        self,
        order_id,
    ):
        '''
        Cancel an order

        :param order_id: required
        :type order_id: str

        :returns: Order

        :raises: DydxAPIError
        '''
        return self._delete(
            '/'.join(['orders', order_id]),
            {},
        )

    def cancel_all_orders(
        self,
        market=None,
    ):
        '''
        Cancel all orders

        :param market: optional
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :returns: Array of orders

        :raises: DydxAPIError
        '''
        params = {'market': market} if market else {}
        return self._delete(
            'orders',
            params,
        )

    def cancel_active_orders(
        self,
        market,
        side=None,
        id=None,
    ):
        '''
        Cancel ActiveOrders

        :param market: required
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param side: optional (required if id is passed in)
        :type side: str in list [
            "BUY",
            "SELL",
        ]

        param id: optional
        :type id: str

        :returns: Array of ActiveOrders

        :raises: DydxAPIError
        '''
        return self._delete(
            'active-orders',
            {
                'market': market,
                'side': side,
                'id': id,
            },
        )

    def get_fills(
        self,
        market=None,
        order_id=None,
        limit=None,
        created_before_or_at=None,
    ):
        '''
        Get fills

        :param market: optional
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param order_id: optional
        :type order_id: str

        :param limit: optional
        :type limit: str

        :param created_before_or_at: optional
        :type created_before_or_at: ISO str

        :returns: Array of fills

        :raises: DydxAPIError
        '''
        return self._get(
            'fills',
            {
                'market': market,
                'orderId': order_id,
                'limit': limit,
                'createdBeforeOrAt': created_before_or_at,
            }
        )

    def get_transfers(
        self,
        transfer_type=None,
        limit=None,
        created_before_or_at=None,
    ):
        '''
        Get transfers

        :param transfer_type: optional
        :type transfer_type: str in list [
            "DEPOSIT",
            "WITHDRAWAL",
            "FAST_WITHDRAWAL",
        ]

        :param limit: optional
        :type limit: str

        :param created_before_or_at: optional
        :type created_before_or_at: ISO str

        :returns: Array of transfers

        :raises: DydxAPIError
        '''
        return self._get(
            'transfers',
            {
                'type': transfer_type,
                'limit': limit,
                'createdBeforeOrAt': created_before_or_at,
            },
        )

    def create_withdrawal(
        self,
        position_id,
        amount,
        asset,
        to_address,
        client_id=None,
        expiration=None,
        expiration_epoch_seconds=None,
        signature=None,
    ):
        '''
        Post a withdrawal

        :param position_id: required
        :type position_id: int or str

        :param amount: required
        :type amount: str

        :param asset: required
        :type asset: str in list [
            "ETH",
            "LINK",
            "BTC",
            "USDC",
            "USDT",
            "USD",
            ...
        ]

        :param client_id: optional
        :type client_id: str

        :param expiration: optional
        :type expiration: ISO str

        :param expiration_epoch_seconds: optional
        :type expiration_epoch_seconds: int

        :param signature: optional
        :type signature: str

        :returns: Transfer

        :raises: DydxAPIError
        '''
        client_id = client_id or random_client_id()
        if bool(expiration) == bool(expiration_epoch_seconds):
            raise ValueError(
                'Exactly one of expiration and expiration_epoch_seconds must '
                'be specified',
            )
        expiration = expiration or epoch_seconds_to_iso(
            expiration_epoch_seconds,
        )
        expiration_epoch_seconds = (
            expiration_epoch_seconds or iso_to_epoch_seconds(expiration)
        )

        if not signature:
            if not self.stark_private_key:
                raise Exception(
                    'No signature provided and client was not' +
                    'initialized with stark_private_key'
                )
            withdrawal_to_sign = SignableWithdrawal(
                network_id=self.network_id,
                position_id=position_id,
                client_id=client_id,
                human_amount=amount,
                expiration_epoch_seconds=expiration_epoch_seconds,
            )
            signature = withdrawal_to_sign.sign(self.stark_private_key)

        params = {
            'amount': amount,
            'asset': asset,
            'expiration': expiration,
            'clientId': client_id,
            'signature': signature,
        }
        return self._post('withdrawals', params)

    def create_transfer(
        self,
        amount,
        position_id,
        receiver_account_id,
        receiver_public_key,
        receiver_position_id,
        client_id=None,
        expiration=None,
        expiration_epoch_seconds=None,
        signature=None,
    ):
        '''
        Create a L2 transfer.

        :param amount: required
        :type amount: str

        :param position_id: required
        :type position_id: int or str

        :param receiver_account_id: required
        :type receiver_account_id: str

        :param receiver_public_key: required
        :type receiver_public_key: str

        :param receiver_position_id: required
        :type receiver_position_id: int or str

        :param client_id: optional
        :type client_id: str

        :param expiration: optional
        :type expiration: ISO str

        :param expiration_epoch_seconds: optional
        :type expiration_epoch_seconds: int

        :param signature: optional
        :type signature: str

        :returns: Transfer

        :raises: DydxAPIError
        '''
        client_id = client_id or random_client_id()

        if bool(expiration) == bool(expiration_epoch_seconds):
            raise ValueError(
                'Exactly one of expiration and expiration_epoch_seconds must '
                'be specified',
            )
        expiration = expiration or epoch_seconds_to_iso(
            expiration_epoch_seconds,
        )
        expiration_epoch_seconds = (
            expiration_epoch_seconds or iso_to_epoch_seconds(expiration)
        )

        transfer_signature = signature
        if not transfer_signature:
            if not self.stark_private_key:
                raise Exception(
                    'No signature provided and client was not'
                    + 'initialized with stark_private_key'
                )
            transfer_to_sign = SignableTransfer(
                network_id=self.network_id,
                sender_position_id=int(position_id),
                receiver_position_id=int(receiver_position_id),
                receiver_public_key=receiver_public_key,
                human_amount=amount,
                client_id=client_id,
                expiration_epoch_seconds=expiration_epoch_seconds,
            )
            transfer_signature = transfer_to_sign.sign(self.stark_private_key)

        params = {
            'amount': amount,
            'receiverAccountId': receiver_account_id,
            'clientId': client_id,
            'signature': transfer_signature,
            'expiration': expiration,
        }
        return self._post('transfers', params)

    def create_fast_withdrawal(
        self,
        position_id,
        credit_asset,
        credit_amount,
        debit_amount,
        to_address,
        lp_position_id,
        lp_stark_public_key,
        slippage_tolerance=None,
        client_id=None,
        expiration=None,
        expiration_epoch_seconds=None,
        signature=None,
    ):
        '''
        Post a fast withdrawal

        :param credit_asset: required
        :type credit_asset: str in list [
            "USDC",
            "USDT",
        ]

        :param position_id: required
        :type position_id: str or int

        :param credit_amount: required
        :type credit_amount: str or int

        :param debit_amount: required
        :type debit_amount: str or int

        :param to_address: required
        :type to_address: str

        :param lp_position_id: required
        :type lp_position_id: str or int

        :param lp_stark_public_key: required
        :type lp_stark_public_key: str

        :param slippage_tolerance: optional
        :type slippage_tolerance: str

        :param client_id: optional
        :type client_id: str

        :param expiration: optional
        :type expiration: ISO str

        :param expiration_epoch_seconds: optional
        :type expiration_epoch_seconds: int

        :param signature: optional
        :type signature: str

        :returns: Transfer

        :raises: DydxAPIError
        '''
        client_id = client_id or random_client_id()
        if bool(expiration) == bool(expiration_epoch_seconds):
            raise ValueError(
                'Exactly one of expiration and expiration_epoch_seconds must '
                'be specified',
            )
        expiration = expiration or epoch_seconds_to_iso(
            expiration_epoch_seconds,
        )
        expiration_epoch_seconds = (
            expiration_epoch_seconds or iso_to_epoch_seconds(expiration)
        )

        if not signature:
            if not self.stark_private_key:
                raise Exception(
                    'No signature provided and client was not' +
                    'initialized with stark_private_key'
                )
            fact = get_transfer_erc20_fact(
                recipient=to_address,
                token_decimals=COLLATERAL_TOKEN_DECIMALS,
                human_amount=credit_amount,
                token_address=(
                    TOKEN_CONTRACTS[COLLATERAL_ASSET][self.network_id]
                ),
                salt=nonce_from_client_id(client_id),
            )
            transfer_to_sign = SignableConditionalTransfer(
                network_id=self.network_id,
                sender_position_id=position_id,
                receiver_position_id=lp_position_id,
                receiver_public_key=lp_stark_public_key,
                fact_registry_address=FACT_REGISTRY_CONTRACT[self.network_id],
                fact=fact,
                human_amount=debit_amount,
                client_id=client_id,
                expiration_epoch_seconds=expiration_epoch_seconds,
            )
            signature = transfer_to_sign.sign(self.stark_private_key)

        params = {
            'creditAsset': credit_asset,
            'creditAmount': credit_amount,
            'debitAmount': debit_amount,
            'slippageTolerance': slippage_tolerance,
            # TODO: Signature verification should work regardless of case.
            'toAddress': to_address.lower(),
            'lpPositionId': lp_position_id,
            'expiration': expiration,
            'clientId': client_id,
            'signature': signature,
        }
        return self._post('fast-withdrawals', params)

    def get_funding_payments(
        self,
        market=None,
        limit=None,
        effective_before_or_at=None,
    ):
        '''
        Get funding payments

        :param market: optional
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param limit: optional
        :type limit: str

        :param effective_before_or_at: optional
        :type effective_before_or_at: ISO str

        :returns: Array of funding payments

        :raises: DydxAPIError
        '''
        return self._get(
            'funding',
            {
                'market': market,
                'limit': limit,
                'effectiveBeforeOrAt': effective_before_or_at,
            },
        )

    def get_historical_pnl(
        self,
        created_before_or_at=None,
        created_on_or_after=None,
    ):
        '''
        Get historical pnl ticks

        :param created_before_or_at: optional
        :type created_before_or_at: ISO str

        :param created_on_or_after: optional
        :type created_on_or_after: ISO str

        :returns: Array of historical pnl ticks

        :raises: DydxAPIError
        '''
        return self._get(
            'historical-pnl',
            {
                'createdBeforeOrAt': created_before_or_at,
                'createdOnOrAfter': created_on_or_after,
            },
        )

    def send_verification_email(
        self,
    ):
        '''
        Send verification email

        :returns: Empty object

        :raises: DydxAPIError
        '''
        return self._put(
            'emails/send-verification-email',
            {},
        )

    def get_trading_rewards(
        self,
        epoch=None,
    ):
        '''
        Get trading rewards

        :param epoch: optional
        :type epoch: int

        :returns: TradingRewards

        :raises: DydxAPIError
        '''
        return self._get(
            'rewards/weight',
            {
                'epoch': epoch,
            },
        )

    def get_liquidity_provider_rewards(
        self,
        epoch=None,
    ):
        '''
        Get liquidity rewards

        :param epoch: optional
        :type epoch: int

        :returns: LiquidityRewards

        :raises: DydxAPIError
        '''
        return self._get(
            'rewards/liquidity',
            {
                'epoch': epoch,
            },
        )

    def get_retroactive_mining_rewards(
        self,
    ):
        '''
        Get retroactive mining rewards

        :returns: RetroactiveMiningRewards

        :raises: DydxAPIError
        '''
        return self._get('rewards/retroactive-mining')

    def request_testnet_tokens(
        self,
    ):
        '''
        Requests tokens on dYdX's staging server.
        NOTE: this will not work on Mainnet/Production.

        :returns: Transfer

        :raises: DydxAPIError
        '''
        if (self.network_id != 3):
            raise ValueError('network_id is not Ropsten')

        return self._post('testnet/tokens', {})

    def get_profile(
        self,
    ):
        '''
        Get Private Profile

        :returns: PrivateProfile

        :raises: DydxAPIError
        '''
        return self._get('profile/private', {})

    def get_user_links(
        self,
    ):
        '''
        Get Active Linked Users

        :returns: UserLinks

        :raises: DydxAPIError
        '''
        return self._get('users/links', {})

    def send_link_request(
        self,
        action,
        address,
    ):
        '''
        Send Link Request Action

        :param action: required
        :type action: str in list [
            "CREATE_SECONDARY_REQUEST",
            "DELETE_SECONDARY_REQUEST",
            "ACCEPT_PRIMARY_REQUEST",
            "REJECT_PRIMARY_REQUEST",
            "REMOVE",
        ]

        :param address: required
        :type address: str

        :returns: {}

        :raises: DydxAPIError
        '''
        return self._post(
            'users/links',
            {
                'action': action,
                'address': address,
            },
        )

    def get_user_pending_link_requests(
        self,
    ):
        '''
        Get Pending Linked User Requests

        :returns: UserLinkRequests

        :raises: DydxAPIError
        '''
        return self._get('users/links/requests', {})

    # ============ Signing ============

    def sign(
        self,
        request_path,
        method,
        iso_timestamp,
        data,
    ):
        message_string = (
            iso_timestamp +
            method +
            request_path +
            (json_stringify(data) if data else '')
        )

        hashed = hmac.new(
            base64.urlsafe_b64decode(
                (self.api_key_credentials['secret']).encode('utf-8'),
            ),
            msg=message_string.encode('utf-8'),
            digestmod=hashlib.sha256,
        )
        return base64.urlsafe_b64encode(hashed.digest()).decode()



from collections import namedtuple
import math

from dydx3.constants import COLLATERAL_ASSET
from dydx3.constants import COLLATERAL_ASSET_ID_BY_NETWORK_ID
from dydx3.starkex.constants import ONE_HOUR_IN_SECONDS
from dydx3.starkex.constants import TRANSFER_FIELD_BIT_LENGTHS
from dydx3.starkex.constants import TRANSFER_PADDING_BITS
from dydx3.starkex.constants import TRANSFER_PREFIX
from dydx3.starkex.constants import TRANSFER_FEE_ASSET_ID
from dydx3.starkex.constants import TRANSFER_MAX_AMOUNT_FEE
from dydx3.starkex.helpers import nonce_from_client_id
from dydx3.starkex.helpers import to_quantums_exact
from dydx3.starkex.signable import Signable
from dydx3.starkex.starkex_resources.proxy import get_hash

StarkwareTransfer = namedtuple(
    'StarkwareTransfer',
    [
        'sender_position_id',
        'receiver_position_id',
        'receiver_public_key',
        'quantums_amount',
        'nounce',
        'expiration_epoch_hours',
    ],
)


class SignableTransfer(Signable):
    """
    Wrapper object to convert a transfer, and hash, sign, and verify its
    signature.
    """

    def __init__(
        self,
        sender_position_id,
        receiver_position_id,
        receiver_public_key,
        human_amount,
        client_id,
        expiration_epoch_seconds,
        network_id,
    ):
        nounce = nonce_from_client_id(client_id)

        # The transfer asset is always the collateral asset.
        quantums_amount = to_quantums_exact(
            human_amount,
            COLLATERAL_ASSET,
        )

        # Convert to a Unix timestamp (in hours).
        expiration_epoch_hours = math.ceil(
            float(expiration_epoch_seconds) / ONE_HOUR_IN_SECONDS,
        )

        receiver_public_key = (
            receiver_public_key
            if isinstance(receiver_public_key, int)
            else int(receiver_public_key, 16)
        )

        message = StarkwareTransfer(
            sender_position_id=int(sender_position_id),
            receiver_position_id=int(receiver_position_id),
            receiver_public_key=receiver_public_key,
            quantums_amount=quantums_amount,
            nounce=nounce,
            expiration_epoch_hours=expiration_epoch_hours
        )

        super(SignableTransfer, self).__init__(network_id, message)

    def to_starkware(self):
        return self._message

    def _calculate_hash(self):
        """Calculate the hash of the Starkware order."""
        # TODO: Check values are in bounds

        asset_ids = get_hash(
            COLLATERAL_ASSET_ID_BY_NETWORK_ID[self.network_id],
            TRANSFER_FEE_ASSET_ID,
        )

        part1 = get_hash(
            asset_ids,
            self._message.receiver_public_key,
        )

        part2 = self._message.sender_position_id
        part2 <<= TRANSFER_FIELD_BIT_LENGTHS['position_id']
        part2 += self._message.receiver_position_id
        part2 <<= TRANSFER_FIELD_BIT_LENGTHS['position_id']
        part2 += self._message.sender_position_id
        part2 <<= TRANSFER_FIELD_BIT_LENGTHS['nonce']
        part2 += self._message.nounce

        part3 = TRANSFER_PREFIX
        part3 <<= TRANSFER_FIELD_BIT_LENGTHS['quantums_amount']
        part3 += self._message.quantums_amount
        part3 <<= TRANSFER_FIELD_BIT_LENGTHS['quantums_amount']
        part3 += TRANSFER_MAX_AMOUNT_FEE
        part3 <<= TRANSFER_FIELD_BIT_LENGTHS['expiration_epoch_hours']
        part3 += self._message.expiration_epoch_hours
        part3 <<= TRANSFER_PADDING_BITS

        return get_hash(
            get_hash(
                part1,
                part2,
            ),
            part3,
        )



from dydx3.eth_signing.eth_prive_action import SignEthPrivateAction
from dydx3.eth_signing.onboarding_action import SignOnboardingAction
from dydx3.eth_signing.signers import SignWithKey
from dydx3.eth_signing.signers import SignWithWeb3



from dydx3.helpers.request_helpers import generate_query_path
from dydx3.helpers.requests import request


class Public(object):

    def __init__(
        self,
        host,
    ):
        self.host = host

    # ============ Request Helpers ============

    def _get(self, request_path, params={}):
        return request(
            generate_query_path(self.host + request_path, params),
            'get',
        )

    def _put(self, endpoint, data):
        return request(
            self.host + '/v3/' + endpoint,
            'put',
            {},
            data,
        )

    # ============ Requests ============

    def check_if_user_exists(self, ethereum_address):
        '''
        Check if user exists

        :param host: required
        :type host: str

        :returns: Bool

        :raises: DydxAPIError
        '''
        uri = '/v3/users/exists'
        return self._get(
            uri,
            {'ethereumAddress': ethereum_address},
        )

    def check_if_username_exists(self, username):
        '''
        Check if username exists

        :param username: required
        :type username: str

        :returns: Bool

        :raises: DydxAPIError
        '''
        uri = '/v3/usernames'
        return self._get(uri, {'username': username})

    def get_markets(self, market=None):
        '''
        Get one or more markets

        :param market: optional
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :returns: Market array

        :raises: DydxAPIError
        '''
        uri = '/v3/markets'
        return self._get(uri, {'market': market})

    def get_orderbook(self, market):
        '''
        Get orderbook for a market

        :param market: required
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :returns: Object containing bid array and ask array of open orders
        for a market

        :raises: DydxAPIError
        '''
        uri = '/'.join(['/v3/orderbook', market])
        return self._get(uri)

    def get_stats(self, market=None, days=None):
        '''
        Get one or more day statistics for a market

        :param market: optional
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param days: optional
        :type days: str in list [
            "1",
            "7",
            "30",
        ]

        :returns: Statistic information for a market, either for all time
        periods or just one.

        :raises: DydxAPIError
        '''
        uri = (
            '/'.join(['/v3/stats', market])
            if market is not None
            else '/v3/stats'
        )

        return self._get(uri, {'days': days})

    def get_trades(self, market, starting_before_or_at=None):
        '''
        Get trades for a market

        :param market: required
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param starting_before_or_at: optional
        :type starting_before_or_at: str

        :returns: Trade array

        :raises: DydxAPIError
        '''
        uri = '/'.join(['/v3/trades', market])
        return self._get(
            uri,
            {'startingBeforeOrAt': starting_before_or_at},
        )

    def get_historical_funding(self, market, effective_before_or_at=None):
        '''
        Get historical funding for a market

        :param market: required
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param effective_before_or_at: optional
        :type effective_before_or_at: str

        :returns: Array of historical funding for a specific market

        :raises: DydxAPIError
        '''
        uri = '/'.join(['/v3/historical-funding', market])
        return self._get(
            uri,
            {'effectiveBeforeOrAt': effective_before_or_at},
        )

    def get_fast_withdrawal(
        self,
        creditAsset=None,
        creditAmount=None,
        debitAmount=None,
    ):
        '''
        Get all fast withdrawal account information

        :param creditAsset: optional
        :type creditAsset: str

        :param creditAmount: optional
        :type creditAmount: str

        :param debitAmount: optional
        :type debitAmount: str

        :returns: All fast withdrawal accounts

        :raises: DydxAPIError
        '''
        uri = '/v3/fast-withdrawals'
        return self._get(
            uri,
            {
                'creditAsset': creditAsset,
                'creditAmount': creditAmount,
                'debitAmount': debitAmount,
            },
        )

    def get_candles(
        self,
        market,
        resolution=None,
        from_iso=None,
        to_iso=None,
        limit=None,
    ):
        '''
        Get Candles

        :param market: required
        :type market: str in list [
            "BTC-USD",
            "ETH-USD",
            "LINK-USD",
            ...
        ]

        :param resolution: optional
        :type resolution: str in list [
            "1DAY",
            "4HOURS"
            "1HOUR",
            "30MINS",
            "15MINS",
            "5MINS",
            "1MIN",
        ]

        :param from_iso: optional
        :type from_iso: str

        :param to_iso: optional
        :type to_iso: str

        :param limit: optional
        :type limit: str

        :returns: Array of candles

        :raises: DydxAPIError
        '''
        uri = '/'.join(['/v3/candles', market])
        return self._get(
            uri,
            {
                'resolution': resolution,
                'fromISO': from_iso,
                'toISO': to_iso,
                'limit': limit,
            },
        )

    def get_time(self):
        '''
        Get api server time as iso and as epoch in seconds with MS

        :returns: ISO string and Epoch number in seconds with MS of server time

        :raises: DydxAPIError
        '''
        uri = '/v3/time'
        return self._get(uri)

    def verify_email(
        self,
        token,
    ):
        '''
        Verify email with token

        :param token: required
        :type token: string

        :returns: empty object

        :raises: DydxAPIError
        '''
        return self._put(
            'emails/verify-email',
            {
                'token': token,
            }
        )

    def get_public_retroactive_mining_rewards(
        self,
        ethereum_address,
    ):
        '''
        Get public retroactive mining rewards

        :param ethereumAddress: required
        :type ethereumAddress: str

        :returns: PublicRetroactiveMiningRewards

        :raises: DydxAPIError
        '''
        return self._get(
            '/v3/rewards/public-retroactive-mining',
            {
                'ethereumAddress': ethereum_address,
            },
        )

    def get_config(self):
        '''
        Get global config variables for the exchange as a whole.
        This includes (but is not limited to) details on the exchange,
        including addresses, fees, transfers, and rate limits.

        :returns: GlobalConfigVariables

        :raises: DydxAPIError
        '''
        return self._get('/v3/config')

    def get_insurance_fund_balance(self):
        '''
        Get the balance of the dYdX insurance fund

        :returns: Balance of the dYdX insurance fund in USD

        :raises: DydxAPIError
        '''
        return self._get('/v3/insurance-fund/balance')

    def get_profile(
        self,
        publicId,
    ):
        '''
        Get Public Profile

        :param publicId: required
        :type publicId: str

        :returns: PublicProfile

        :raises: DydxAPIError
        '''
        uri = '/'.join(['/v3/profile', publicId])
        return self._get(uri)

    def get_historical_leaderboard_pnls(
        self,
        period,
        limit=None,
    ):
        '''
        Get Historical Leaderboard Pnls

        :param period: required
        :type period: str
        :type period: str in list [
            "LEAGUES",
            "DAILY",
            "DAILY_COMPETITION",
            ...
        ]

        :param limit: optional
        :type limit: str

        :returns: HistoricalLeaderboardPnl

        :raises: DydxAPIError
        '''
        uri = '/'.join(['/v3/accounts/historical-leaderboard-pnls', period])
        return self._get(
            uri,
            {
                'limit': limit,
            }
        )



from web3 import Web3

from dydx3.eth_signing import util
from dydx3.eth_signing.sign_off_chain_action import SignOffChainAction

EIP712_ETH_PRIVATE_ACTION_STRUCT_STRING = [
    {'type': 'string', 'name': 'method'},
    {'type': 'string', 'name': 'requestPath'},
    {'type': 'string', 'name': 'body'},
    {'type': 'string', 'name': 'timestamp'},
]
EIP712_ETH_PRIVATE_ACTION_STRUCT_STRING_STRING = (
    'dYdX(' +
    'string method,' +
    'string requestPath,' +
    'string body,' +
    'string timestamp' +
    ')'
)
EIP712_STRUCT_NAME = 'dYdX'


class SignEthPrivateAction(SignOffChainAction):

    def get_eip712_struct(self):
        return EIP712_ETH_PRIVATE_ACTION_STRUCT_STRING

    def get_eip712_struct_name(self):
        return EIP712_STRUCT_NAME

    def get_eip712_message(
        self,
        method,
        request_path,
        body,
        timestamp,
    ):
        return super(SignEthPrivateAction, self).get_eip712_message(
            method=method,
            requestPath=request_path,
            body=body,
            timestamp=timestamp,
        )

    def get_hash(
        self,
        method,
        request_path,
        body,
        timestamp,
    ):
        data = [
            [
                'bytes32',
                'bytes32',
                'bytes32',
                'bytes32',
                'bytes32',
            ],
            [
                util.hash_string(
                    EIP712_ETH_PRIVATE_ACTION_STRUCT_STRING_STRING,
                ),
                util.hash_string(method),
                util.hash_string(request_path),
                util.hash_string(body),
                util.hash_string(timestamp),
            ],
        ]
        struct_hash = Web3.solidityKeccak(*data)
        return self.get_eip712_hash(struct_hash)



import os

from dydx3 import Client
from dydx3.constants import MARKET_BTC_USD
from dydx3.constants import MARKET_STATISTIC_DAY_ONE

from tests.constants import DEFAULT_HOST

ADDRESS_1 = '0x90F8bf6A479f320ead074411a4B0e7944Ea8c9C0'
API_HOST = os.environ.get('V3_API_HOST', DEFAULT_HOST)


class TestPublic():

    def test_check_if_user_exists(self):
        public = Client(API_HOST).public
        resp = public.check_if_user_exists(ADDRESS_1)
        expected_data = {
            'exists': False,
            'contractAddress': '',
            'isProxySigner': False
        }
        assert resp.data == expected_data
        assert resp.headers != {}

    def test_check_if_username_exists(self):
        public = Client(API_HOST).public
        resp = public.check_if_username_exists('foo')
        assert resp.data == {'exists': False}
        assert resp.headers != {}

    def test_get_markets(self):
        public = Client(API_HOST).public
        resp = public.get_markets()
        assert resp.data != {}
        assert resp.headers != {}

    def test_get_orderbook(self):
        public = Client(API_HOST).public
        resp = public.get_orderbook(MARKET_BTC_USD)
        assert resp.data != {}
        assert resp.headers != {}

    def test_get_stats(self):
        public = Client(API_HOST).public
        resp = public.get_stats(
            MARKET_BTC_USD,
            MARKET_STATISTIC_DAY_ONE,
        )
        assert resp.data != {}
        assert resp.headers != {}

    def test_get_trades(self):
        public = Client(API_HOST).public
        resp = public.get_trades(MARKET_BTC_USD)
        assert resp.data != {}
        assert resp.headers != {}

    def test_get_historical_funding(self):
        public = Client(API_HOST).public
        resp = public.get_historical_funding(MARKET_BTC_USD)
        assert resp.data != {}
        assert resp.headers != {}

    def test_get_candles(self):
        public = Client(API_HOST).public
        resp = public.get_candles(MARKET_BTC_USD)
        assert resp.data != {}
        assert resp.headers != {}

    def test_get_fast_withdrawal(self):
        public = Client(API_HOST).public
        resp = public.get_fast_withdrawal()
        assert resp.data != {}
        assert resp.headers != {}

    def test_verify_email(self):
        try:
            public = Client(API_HOST).public
            public.verify_email('token')
        except Exception as e:
            # No userId gotten with token: token so no verification
            # has occurred
            assert e.status_code == 400

    def test_public_retroactive_mining(self):
        public = Client(API_HOST).public
        resp = public.get_public_retroactive_mining_rewards(ADDRESS_1)
        assert resp.data != {}
        assert resp.headers != {}



from typing import Optional, Union

from dydx3.starkex.starkex_resources.cpp_signature import check_cpp_lib_path
from dydx3.starkex.starkex_resources.cpp_signature import cpp_hash
from dydx3.starkex.starkex_resources.cpp_signature import cpp_verify
from dydx3.starkex.starkex_resources.python_signature import ECPoint
from dydx3.starkex.starkex_resources.python_signature import ECSignature
from dydx3.starkex.starkex_resources.python_signature import py_pedersen_hash
from dydx3.starkex.starkex_resources.python_signature import py_sign
from dydx3.starkex.starkex_resources.python_signature import py_verify


def sign(
    msg_hash: int,
    priv_key: int,
    seed: Optional[int] = None,
) -> ECSignature:
    # Note: cpp_sign() is not optimized and is currently slower than py_sign().
    #       So always use py_sign() for now.
    return py_sign(msg_hash=msg_hash, priv_key=priv_key, seed=seed)


def verify(
    msg_hash: int,
    r: int,
    s: int,
    public_key: Union[int, ECPoint],
) -> bool:
    if check_cpp_lib_path():
        return cpp_verify(msg_hash=msg_hash, r=r, s=s, stark_key=public_key)

    return py_verify(msg_hash=msg_hash, r=r, s=s, public_key=public_key)


def get_hash(*elements: int) -> int:
    if check_cpp_lib_path():
        return cpp_hash(*elements)

    return py_pedersen_hash(*elements)



from dydx3.constants import NETWORK_ID_ROPSTEN
from dydx3.helpers.request_helpers import iso_to_epoch_seconds
from dydx3.starkex.conditional_transfer import SignableConditionalTransfer

MOCK_PUBLIC_KEY = (
    '3b865a18323b8d147a12c556bfb1d502516c325b1477a23ba6c77af31f020fd'
)
MOCK_PRIVATE_KEY = (
    '58c7d5a90b1776bde86ebac077e053ed85b0f7164f53b080304a531947f46e3'
)
MOCK_SIGNATURE = (
    '04814c5d3501863134108802cab5d12df4b959654332103b837252549d24e9a6'
    + '06bc01225e9f1690b08b63de2a3b179fb2927d4564b3440bbb0da4c37caf597e'
)

# Mock conditional transfer params.
CONDITIONAL_TRANSFER_PARAMS = {
    "network_id": NETWORK_ID_ROPSTEN,
    'sender_position_id': 12345,
    'receiver_position_id': 67890,
    'receiver_public_key': (
        '05135ef87716b0faecec3ba672d145a6daad0aa46437c365d490022115aba674'
    ),
    'fact_registry_address': '0x12aa12aa12aa12aa12aa12aa12aa12aa12aa12aa',
    'fact': bytes.fromhex(
        '12ff12ff12ff12ff12ff12ff12ff12ff12ff12ff12ff12ff12ff12ff12ff12ff'
    ),
    'human_amount': '49.478023',
    'expiration_epoch_seconds': iso_to_epoch_seconds(
        '2020-09-17T04:15:55.028Z',
    ),
    'client_id': (
        'This is an ID that the client came up with to describe this transfer'
    ),
}


class TestConditionalTransfer():

    def test_sign_conditional_transfer(self):
        transfer = SignableConditionalTransfer(**CONDITIONAL_TRANSFER_PARAMS)
        signature = transfer.sign(MOCK_PRIVATE_KEY)
        assert signature == MOCK_SIGNATURE

    def test_verify_signature(self):
        transfer = SignableConditionalTransfer(**CONDITIONAL_TRANSFER_PARAMS)
        assert transfer.verify_signature(MOCK_SIGNATURE, MOCK_PUBLIC_KEY)

    def test_starkware_representation(self):
        transfer = SignableConditionalTransfer(**CONDITIONAL_TRANSFER_PARAMS)
        starkware_transfer = transfer.to_starkware()
        assert starkware_transfer.quantums_amount == 49478023

        # Order expiration should be rounded up and should have a buffer added.
        assert starkware_transfer.expiration_epoch_hours == 444533



import os
import time

from web3 import Web3

from dydx3 import Client
from dydx3 import DydxApiError
from dydx3 import SignableWithdrawal
from dydx3 import constants
from dydx3 import generate_private_key_hex_unsafe
from dydx3 import private_key_to_public_key_pair_hex
from dydx3.helpers.request_helpers import random_client_id

from tests.constants import DEFAULT_HOST
from tests.constants import DEFAULT_NETWORK_ID
from tests.constants import SEVEN_DAYS_S

HOST = os.environ.get('V3_API_HOST', DEFAULT_HOST)
NETWORK_ID = int(os.environ.get('NETWORK_ID', DEFAULT_NETWORK_ID))


class TestAuthLevels():

    def test_public(self):
        client = Client(
            host=HOST,
            network_id=NETWORK_ID,
        )
        client.public.get_markets()

    def test_private_with_private_keys(self):
        # Generate STARK keys and Ethhereum account.
        stark_private_key = generate_private_key_hex_unsafe()
        eth_account = Web3(None).eth.account.create()

        # Get public key.
        stark_public_key, stark_public_key_y_coordinate = (
            private_key_to_public_key_pair_hex(stark_private_key)
        )

        # Onboard the user.
        Client(
            host=HOST,
            network_id=NETWORK_ID,
            eth_private_key=eth_account.key,
        ).onboarding.create_user(
            stark_public_key=stark_public_key,
            stark_public_key_y_coordinate=stark_public_key_y_coordinate,
        )

        # Create a second client WITHOUT eth_private_key.
        client = Client(
            host=HOST,
            network_id=NETWORK_ID,
            stark_private_key=stark_private_key,
        )

        # Get the primary account.
        get_account_result = client.private.get_account(
            ethereum_address=eth_account.address,
        )
        account = get_account_result['account']

        # Initiate a regular (slow) withdrawal.
        #
        # Expect signature validation to pass, although the collateralization
        # check will fail.
        expected_error = (
            'Withdrawal would put account under collateralization minumum'
        )
        expiration_epoch_seconds = time.time() + SEVEN_DAYS_S + 60
        try:
            client.private.create_withdrawal(
                position_id=account['positionId'],
                amount='1',
                asset=constants.ASSET_USDC,
                to_address=eth_account.address,
                expiration_epoch_seconds=expiration_epoch_seconds,
            )
        except DydxApiError as e:
            if expected_error not in str(e):
                raise

    def test_private_without_stark_private_key(self):
        # Generate STARK keys and Ethhereum account.
        stark_private_key = generate_private_key_hex_unsafe()
        eth_account = Web3(None).eth.account.create()

        # Get public key.
        stark_public_key, stark_public_key_y_coordinate = (
            private_key_to_public_key_pair_hex(stark_private_key)
        )

        # Onboard the user.
        Client(
            host=HOST,
            network_id=NETWORK_ID,
            eth_private_key=eth_account.key,
        ).onboarding.create_user(
            stark_public_key=stark_public_key,
            stark_public_key_y_coordinate=stark_public_key_y_coordinate,
        )

        # Create a second client WITHOUT eth_private_key or stark_private_key.
        client = Client(
            host=HOST,
            network_id=NETWORK_ID,
        )

        # Get the primary account.
        get_account_result = client.private.get_account(
            ethereum_address=eth_account.address,
        )
        account = get_account_result['account']

        # Sign a withdrawal.
        client_id = random_client_id()
        expiration_epoch_seconds = time.time() + SEVEN_DAYS_S + 60
        signable_withdrawal = SignableWithdrawal(
            network_id=NETWORK_ID,
            position_id=account['positionId'],
            client_id=client_id,
            human_amount='1',
            expiration_epoch_seconds=expiration_epoch_seconds,
        )
        signature = signable_withdrawal.sign(stark_private_key)

        # Initiate a regular (slow) withdrawal.
        #
        # Expect signature validation to pass, although the collateralization
        # check will fail.
        expected_error = (
            'Withdrawal would put account under collateralization minumum'
        )
        try:
            client.private.create_withdrawal(
                position_id=account['positionId'],
                amount='1',
                asset=constants.ASSET_USDC,
                to_address=eth_account.address,
                client_id=client_id,
                expiration_epoch_seconds=expiration_epoch_seconds,
                signature=signature,
            )
        except DydxApiError as e:
            if expected_error not in str(e):
                raise

    def test_onboard_with_private_keys(self):
        # Generate keys.
        stark_private_key = generate_private_key_hex_unsafe()
        eth_private_key = Web3(None).eth.account.create().key

        # Create client WITH private keys.
        client = Client(
            host=HOST,
            network_id=NETWORK_ID,
            stark_private_key=stark_private_key,
            eth_private_key=eth_private_key,
        )

        # Onboard the user.
        client.onboarding.create_user()

        # Register and then revoke a second API key.
        client.eth_private.create_api_key()
        client.private.get_api_keys()

        # TODO
        # client.eth_private.delete_api_key(api_public_key_2)

    def test_onboard_with_web3_provider(self):
        # Generate private key.
        stark_private_key = generate_private_key_hex_unsafe()

        # Get public key.
        stark_public_key, stark_public_key_y_coordinate = (
            private_key_to_public_key_pair_hex(stark_private_key)
        )

        # Get account address from local Ethereum node.
        ethereum_address = Web3().eth.accounts[0]

        # Create client WITHOUT any private keys.
        client = Client(
            host=HOST,
            network_id=NETWORK_ID,
            web3_provider=Web3.HTTPProvider('http://localhost:8545'),
        )

        # Onboard the user.
        try:
            client.onboarding.create_user(
                ethereum_address=ethereum_address,
                stark_public_key=stark_public_key,
                stark_public_key_y_coordinate=stark_public_key_y_coordinate,
            )

        # If the Ethereum address was already onboarded, ignore the error.
        except DydxApiError:
            pass

        # Register and then revoke a second API key.
        client.eth_private.create_api_key(
            ethereum_address=ethereum_address,
        )
        client.private.get_api_keys()
        client.eth_private.delete_api_key(
            api_key=client.api_key_credentials['key'],
            ethereum_address=ethereum_address,
        )

    def test_onboard_with_web3_default_account(self):
        # Generate private key.
        stark_private_key = generate_private_key_hex_unsafe()

        # Get public key.
        stark_public_key, stark_public_key_y_coordinate = (
            private_key_to_public_key_pair_hex(stark_private_key)
        )

        # Connect to local Ethereum node.
        web3 = Web3()
        web3.eth.defaultAccount = web3.eth.accounts[1]

        # Create client WITHOUT any private keys.
        client = Client(
            host=HOST,
            network_id=NETWORK_ID,
            web3=web3,
        )

        # Onboard the user.
        try:
            client.onboarding.create_user(
                stark_public_key=stark_public_key,
                stark_public_key_y_coordinate=stark_public_key_y_coordinate,
            )

        # If the Ethereum address was already onboarded, ignore the error.
        except DydxApiError:
            pass

        # Register and then revoke a second API key.
        client.eth_private.create_api_key()
        client.private.get_api_keys()
        client.eth_private.delete_api_key(
            api_key=client.api_key_credentials['key'],
        )



import time


class TimedOutWaitingForCondition(Exception):

    def __init__(self, last_value, expected_value):
        self.last_value = last_value
        self.expected_value = expected_value


def wait_for_condition(fn, expected_value, timeout_s, interval_s=1):
    start = time.time()
    result = fn()
    while result != expected_value:
        if time.time() - start > timeout_s:
            raise TimedOutWaitingForCondition(result, expected_value)
        time.sleep(interval_s)
        if time.time() - start > timeout_s:
            raise TimedOutWaitingForCondition(result, expected_value)
        result = fn()
    return result



import decimal
import math

from collections import namedtuple

from dydx3.constants import COLLATERAL_ASSET
from dydx3.constants import COLLATERAL_ASSET_ID_BY_NETWORK_ID
from dydx3.constants import ORDER_SIDE_BUY
from dydx3.constants import SYNTHETIC_ASSET_ID_MAP
from dydx3.constants import SYNTHETIC_ASSET_MAP
from dydx3.starkex.constants import ONE_HOUR_IN_SECONDS
from dydx3.starkex.constants import ORDER_FIELD_BIT_LENGTHS
from dydx3.starkex.constants import ORDER_PADDING_BITS
from dydx3.starkex.constants import ORDER_PREFIX
from dydx3.starkex.constants import ORDER_SIGNATURE_EXPIRATION_BUFFER_HOURS
from dydx3.starkex.helpers import nonce_from_client_id
from dydx3.starkex.helpers import to_quantums_exact
from dydx3.starkex.helpers import to_quantums_round_down
from dydx3.starkex.helpers import to_quantums_round_up
from dydx3.starkex.signable import Signable
from dydx3.starkex.starkex_resources.proxy import get_hash

DECIMAL_CONTEXT_ROUND_DOWN = decimal.Context(rounding=decimal.ROUND_DOWN)
DECIMAL_CONTEXT_ROUND_UP = decimal.Context(rounding=decimal.ROUND_UP)

StarkwareOrder = namedtuple(
    'StarkwareOrder',
    [
        'order_type',
        'asset_id_synthetic',
        'asset_id_collateral',
        'asset_id_fee',
        'quantums_amount_synthetic',
        'quantums_amount_collateral',
        'quantums_amount_fee',
        'is_buying_synthetic',
        'position_id',
        'nonce',
        'expiration_epoch_hours',
    ],
)


class SignableOrder(Signable):

    def __init__(
        self,
        network_id,
        market,
        side,
        position_id,
        human_size,
        human_price,
        limit_fee,
        client_id,
        expiration_epoch_seconds,
    ):
        synthetic_asset = SYNTHETIC_ASSET_MAP[market]
        synthetic_asset_id = SYNTHETIC_ASSET_ID_MAP[synthetic_asset]
        collateral_asset_id = COLLATERAL_ASSET_ID_BY_NETWORK_ID[network_id]
        is_buying_synthetic = side == ORDER_SIDE_BUY
        quantums_amount_synthetic = to_quantums_exact(
            human_size,
            synthetic_asset,
        )

        # Note: By creating the decimals outside the context and then
        # multiplying within the context, we ensure rounding does not occur
        # until after the multiplication is computed with full precision.
        if is_buying_synthetic:
            human_cost = DECIMAL_CONTEXT_ROUND_UP.multiply(
                decimal.Decimal(human_size),
                decimal.Decimal(human_price)
            )
            quantums_amount_collateral = to_quantums_round_up(
                human_cost,
                COLLATERAL_ASSET,
            )
        else:
            human_cost = DECIMAL_CONTEXT_ROUND_DOWN.multiply(
                decimal.Decimal(human_size),
                decimal.Decimal(human_price)
            )
            quantums_amount_collateral = to_quantums_round_down(
                human_cost,
                COLLATERAL_ASSET,
            )

        # The limitFee is a fraction, e.g. 0.01 is a 1 % fee.
        # It is always paid in the collateral asset.
        # Constrain the limit fee to six decimals of precision.
        # The final fee amount must be rounded up.
        limit_fee_rounded = DECIMAL_CONTEXT_ROUND_DOWN.quantize(
            decimal.Decimal(limit_fee),
            decimal.Decimal('0.000001'),
        )
        quantums_amount_fee_decimal = DECIMAL_CONTEXT_ROUND_UP.multiply(
            limit_fee_rounded,
            quantums_amount_collateral,
        ).to_integral_value(context=DECIMAL_CONTEXT_ROUND_UP)

        # Orders may have a short time-to-live on the orderbook, but we need
        # to ensure their signatures are valid by the time they reach the
        # blockchain. Therefore, we enforce that the signed expiration includes
        # a buffer relative to the expiration timestamp sent to the dYdX API.
        expiration_epoch_hours = math.ceil(
            float(expiration_epoch_seconds) / ONE_HOUR_IN_SECONDS,
        ) + ORDER_SIGNATURE_EXPIRATION_BUFFER_HOURS

        message = StarkwareOrder(
            order_type='LIMIT_ORDER_WITH_FEES',
            asset_id_synthetic=synthetic_asset_id,
            asset_id_collateral=collateral_asset_id,
            asset_id_fee=collateral_asset_id,
            quantums_amount_synthetic=quantums_amount_synthetic,
            quantums_amount_collateral=quantums_amount_collateral,
            quantums_amount_fee=int(quantums_amount_fee_decimal),
            is_buying_synthetic=is_buying_synthetic,
            position_id=int(position_id),
            nonce=nonce_from_client_id(client_id),
            expiration_epoch_hours=expiration_epoch_hours,
        )
        super(SignableOrder, self).__init__(network_id, message)

    def to_starkware(self):
        return self._message

    def _calculate_hash(self):
        """Calculate the hash of the Starkware order."""

        # TODO: Check values are in bounds

        if self._message.is_buying_synthetic:
            asset_id_sell = self._message.asset_id_collateral
            asset_id_buy = self._message.asset_id_synthetic
            quantums_amount_sell = self._message.quantums_amount_collateral
            quantums_amount_buy = self._message.quantums_amount_synthetic
        else:
            asset_id_sell = self._message.asset_id_synthetic
            asset_id_buy = self._message.asset_id_collateral
            quantums_amount_sell = self._message.quantums_amount_synthetic
            quantums_amount_buy = self._message.quantums_amount_collateral

        part_1 = quantums_amount_sell
        part_1 <<= ORDER_FIELD_BIT_LENGTHS['quantums_amount']
        part_1 += quantums_amount_buy
        part_1 <<= ORDER_FIELD_BIT_LENGTHS['quantums_amount']
        part_1 += self._message.quantums_amount_fee
        part_1 <<= ORDER_FIELD_BIT_LENGTHS['nonce']
        part_1 += self._message.nonce

        part_2 = ORDER_PREFIX
        for _ in range(3):
            part_2 <<= ORDER_FIELD_BIT_LENGTHS['position_id']
            part_2 += self._message.position_id
        part_2 <<= ORDER_FIELD_BIT_LENGTHS['expiration_epoch_hours']
        part_2 += self._message.expiration_epoch_hours
        part_2 <<= ORDER_PADDING_BITS

        assets_hash = get_hash(
            get_hash(
                asset_id_sell,
                asset_id_buy,
            ),
            self._message.asset_id_fee,
        )
        return get_hash(
            get_hash(
                assets_hash,
                part_1,
            ),
            part_2,
        )



from dydx3.helpers.request_helpers import generate_now_iso
from dydx3.helpers.request_helpers import generate_query_path
from dydx3.helpers.request_helpers import json_stringify
from dydx3.eth_signing import SignEthPrivateAction
from dydx3.helpers.requests import request


class EthPrivate(object):
    """Module for adding/deleting API keys and recovery."""

    def __init__(
        self,
        host,
        eth_signer,
        network_id,
        default_address,
    ):
        self.host = host
        self.default_address = default_address

        self.signer = SignEthPrivateAction(eth_signer, network_id)

    # ============ Request Helpers ============

    def _request(
        self,
        method,
        endpoint,
        opt_ethereum_address,
        data={}
    ):
        ethereum_address = opt_ethereum_address or self.default_address

        request_path = '/'.join(['/v3', endpoint])
        timestamp = generate_now_iso()
        signature = self.signer.sign(
            ethereum_address,
            method=method.upper(),
            request_path=request_path,
            body=json_stringify(data) if data else '{}',
            timestamp=timestamp,
        )

        return request(
            self.host + request_path,
            method,
            {
                'DYDX-SIGNATURE': signature,
                'DYDX-TIMESTAMP': timestamp,
                'DYDX-ETHEREUM-ADDRESS': ethereum_address,
            },
            data,
        )

    def _post(
        self,
        endpoint,
        opt_ethereum_address,
    ):
        return self._request(
            'post',
            endpoint,
            opt_ethereum_address,
        )

    def _delete(
        self,
        endpoint,
        opt_ethereum_address,
        params={},
    ):
        url = generate_query_path(endpoint, params)
        return self._request(
            'delete',
            url,
            opt_ethereum_address,
        )

    def _get(
        self,
        endpoint,
        opt_ethereum_address,
        params={},
    ):
        url = generate_query_path(endpoint, params)
        return self._request(
            'get',
            url,
            opt_ethereum_address,
        )

# ============ Requests ============

    def create_api_key(
        self,
        ethereum_address=None,
    ):
        '''
        Register an API key.

        :param ethereum_address: optional
        :type ethereum_address: str

        :returns: Object containing an apiKey

        :raises: DydxAPIError
        '''
        return self._post(
            'api-keys',
            ethereum_address,
        )

    def delete_api_key(
        self,
        api_key,
        ethereum_address=None,
    ):
        '''
        Delete an API key.

        :param api_key: required
        :type api_key: str

        :param ethereum_address: optional
        :type ethereum_address: str

        :returns: None

        :raises: DydxAPIError
        '''
        return self._delete(
            'api-keys',
            ethereum_address,
            {
                'apiKey': api_key,
            },
        )

    def recovery(
        self,
        ethereum_address=None
    ):
        '''
        This is for if you can't recover your starkKey or apiKey and need an
        additional way to get your starkKey and balance on our exchange,
        both of which are needed to call the L1 solidity function needed to
        recover your funds.

        :param ethereum_address: optional
        :type ethereum_address: str

        :returns: {
            starkKey: str,
            positionId: str,
            equity: str,
            freeCollateral: str,
            quoteBalance: str,
            positions: array of open positions
        }

        :raises: DydxAPIError
        '''
        return self._get(
            'recovery',
            ethereum_address,
        )



from dydx3.constants import NETWORK_ID_ROPSTEN
from dydx3.helpers.request_helpers import iso_to_epoch_seconds
from dydx3.starkex.transfer import SignableTransfer

MOCK_PUBLIC_KEY = (
    '3b865a18323b8d147a12c556bfb1d502516c325b1477a23ba6c77af31f020fd'
)
MOCK_PRIVATE_KEY = (
    '58c7d5a90b1776bde86ebac077e053ed85b0f7164f53b080304a531947f46e3'
)
MOCK_SIGNATURE = (
    '06b72146028a7f0092557a3a04e9916bd4ae1fba0a4bd92670ef80e2293f7386'
    + '04c0918a7a8e622e463d40f24984c23fd8bab2cd32980676ba666f55c6efeaf3'
)

# Mock transfer params.
TRANSFER_PARAMS = {
    'sender_position_id': 12345,
    'receiver_position_id': 67890,
    'receiver_public_key': (
        '05135ef87716b0faecec3ba672d145a6daad0aa46437c365d490022115aba674'
    ),
    'human_amount': '49.478023',
    'expiration_epoch_seconds': iso_to_epoch_seconds(
        '2020-09-17T04:15:55.028Z',
    ),
    'client_id': (
        'This is an ID that the client came up with to describe this transfer'
    ),
}


class TestTransfer():

    def test_sign_transfer(self):
        transfer = SignableTransfer(
            **TRANSFER_PARAMS, network_id=NETWORK_ID_ROPSTEN)
        signature = transfer.sign(MOCK_PRIVATE_KEY)
        assert signature == MOCK_SIGNATURE

    def test_sign_transfer_different_client_id(self):
        alternative_transfer_params = {**TRANSFER_PARAMS}
        alternative_transfer_params['client_id'] += '!'

        transfer = SignableTransfer(
            **alternative_transfer_params,
            network_id=NETWORK_ID_ROPSTEN
        )
        signature = transfer.sign(MOCK_PRIVATE_KEY)
        assert signature != MOCK_SIGNATURE

    def test_sign_transfer_different_receiver_position_id(self):
        alternative_transfer_params = {**TRANSFER_PARAMS}
        alternative_transfer_params['receiver_position_id'] += 1

        transfer = SignableTransfer(
            **alternative_transfer_params,
            network_id=NETWORK_ID_ROPSTEN
        )
        signature = transfer.sign(MOCK_PRIVATE_KEY)
        assert signature != MOCK_SIGNATURE

    def test_verify_signature(self):
        transfer = SignableTransfer(
            **TRANSFER_PARAMS, network_id=NETWORK_ID_ROPSTEN)
        assert transfer.verify_signature(MOCK_SIGNATURE, MOCK_PUBLIC_KEY)

    def test_starkware_representation(self):
        transfer = SignableTransfer(
            **TRANSFER_PARAMS, network_id=NETWORK_ID_ROPSTEN)
        starkware_transfer = transfer.to_starkware()
        assert starkware_transfer.quantums_amount == 49478023

        # Order expiration should be rounded up and should have a buffer added.
        assert starkware_transfer.expiration_epoch_hours == 444533



from dydx3.constants import (
    SYNTHETIC_ASSET_MAP,
    SYNTHETIC_ASSET_ID_MAP,
    ASSET_RESOLUTION,
    COLLATERAL_ASSET,
)


class TestConstants():
    def test_constants_have_regular_structure(self):
        for market, asset in SYNTHETIC_ASSET_MAP.items():
            market_parts = market.split('-')
            base_token, quote_token = market_parts
            assert base_token == asset
            assert quote_token == 'USD'
            assert len(market_parts) == 2

        assert list(SYNTHETIC_ASSET_MAP.values()) \
            == list(SYNTHETIC_ASSET_ID_MAP.keys())

        assets = [x for x in ASSET_RESOLUTION.keys() if x != COLLATERAL_ASSET]
        assert assets == list(SYNTHETIC_ASSET_MAP.values())



from web3 import Web3

from dydx3.eth_signing import util

DOMAIN = 'dYdX'
VERSION = '1.0'
EIP712_DOMAIN_STRING_NO_CONTRACT = (
    'EIP712Domain(' +
    'string name,' +
    'string version,' +
    'uint256 chainId' +
    ')'
)


class SignOffChainAction(object):

    def __init__(self, signer, network_id):
        self.signer = signer
        self.network_id = network_id

    def get_hash(self, **message):
        raise NotImplementedError

    def get_eip712_struct(self):
        raise NotImplementedError

    def get_eip712_struct_name(self):
        raise NotImplementedError

    def sign(
        self,
        signer_address,
        **message,
    ):
        eip712_message = self.get_eip712_message(**message)
        message_hash = self.get_hash(**message)
        typed_signature = self.signer.sign(
            eip712_message,
            message_hash,
            signer_address,
        )
        return typed_signature

    def verify(
        self,
        typed_signature,
        expected_signer_address,
        **message,
    ):
        message_hash = self.get_hash(**message)
        signer = util.ec_recover_typed_signature(message_hash, typed_signature)
        return util.addresses_are_equal(signer, expected_signer_address)

    def get_eip712_message(
        self,
        **message,
    ):
        struct_name = self.get_eip712_struct_name()
        return {
            'types': {
                'EIP712Domain': [
                    {
                        'name': 'name',
                        'type': 'string',
                    },
                    {
                        'name': 'version',
                        'type': 'string',
                    },
                    {
                        'name': 'chainId',
                        'type': 'uint256',
                    },
                ],
                struct_name: self.get_eip712_struct(),
            },
            'domain': {
                'name': DOMAIN,
                'version': VERSION,
                'chainId': self.network_id,
            },
            'primaryType': struct_name,
            'message': message,
        }

    def get_eip712_hash(self, struct_hash):
        return Web3.solidityKeccak(
            [
                'bytes2',
                'bytes32',
                'bytes32',
            ],
            [
                '0x1901',
                self.get_domain_hash(),
                struct_hash,
            ]
        )

    def get_domain_hash(self):
        return Web3.solidityKeccak(
            [
                'bytes32',
                'bytes32',
                'bytes32',
                'uint256',
            ],
            [
                util.hash_string(EIP712_DOMAIN_STRING_NO_CONTRACT),
                util.hash_string(DOMAIN),
                util.hash_string(VERSION),
                self.network_id,
            ],
        )



'''Integration test.
This test can be very slow due to on-chain calls.
Run with `pytest -s` to enable print statements.
'''

import re
import os
import time

from web3 import Web3

from dydx3 import Client
from dydx3 import DydxApiError
from dydx3 import constants
from dydx3 import epoch_seconds_to_iso
from dydx3 import generate_private_key_hex_unsafe
from dydx3 import private_key_to_public_key_pair_hex

from tests.constants import DEFAULT_HOST
from tests.constants import DEFAULT_NETWORK_ID
from tests.constants import SEVEN_DAYS_S

from integration_tests.util import wait_for_condition

HOST = os.environ.get('V3_API_HOST', DEFAULT_HOST)
NETWORK_ID = os.environ.get('NETWORK_ID', DEFAULT_NETWORK_ID)
LP_POSITION_ID = os.environ.get('LP_POSITION_ID', '2')
LP_PUBLIC_KEY = os.environ.get(
    'LP_PUBLIC_KEY',
    '04a9ecd28a67407c3cff8937f329ca24fd631b1d9ca2b9f2df47c7ebf72bf0b0',
)


class TestIntegration():

    def test_integration_without_funds(self):
        # Create an Ethereum account and STARK keys for the new user.
        web3_account = Web3(None).eth.account.create()
        ethereum_address = web3_account.address
        stark_private_key = generate_private_key_hex_unsafe()

        # Create client for the new user.
        client = Client(
            host=HOST,
            network_id=NETWORK_ID,
            stark_private_key=stark_private_key,
            web3_account=web3_account,
        )

        # Onboard the user.
        client.onboarding.create_user()

        # Register a new API key.
        client.eth_private.create_api_key()

        # Get the primary account.
        get_account_result = client.private.get_account(
            ethereum_address=ethereum_address,
        )
        account = get_account_result['account']
        assert int(account['starkKey'], 16) == int(client.stark_public_key, 16)

        # Initiate a regular (slow) withdrawal.
        #
        # Expect signature validation to pass, although the collateralization
        # check will fail.
        expected_error = (
            'Withdrawal would put account under collateralization minumum'
        )
        expiration_epoch_seconds = time.time() + SEVEN_DAYS_S + 60
        try:
            client.private.create_withdrawal(
                position_id=account['positionId'],
                amount='1',
                asset=constants.ASSET_USDC,
                to_address=ethereum_address,
                expiration_epoch_seconds=expiration_epoch_seconds,
            )
        except DydxApiError as e:
            if expected_error not in str(e):
                raise

        # Post an order.
        #
        # Expect signature validation to pass, although the collateralization
        # check will fail.
        one_minute_from_now_iso = epoch_seconds_to_iso(time.time() + 60)
        try:
            client.private.create_order(
                position_id=account['positionId'],
                market=constants.MARKET_BTC_USD,
                side=constants.ORDER_SIDE_BUY,
                order_type=constants.ORDER_TYPE_LIMIT,
                post_only=False,
                size='10',
                price='1000',
                limit_fee='0.1',
                expiration=one_minute_from_now_iso,
            )
        except DydxApiError as e:
            if expected_error not in str(e):
                raise

    def test_integration(self):
        source_private_key = os.environ.get('TEST_SOURCE_PRIVATE_KEY')
        if source_private_key is None:
            raise ValueError('TEST_SOURCE_PRIVATE_KEY must be set')

        web3_provider = os.environ.get('TEST_WEB3_PROVIDER_URL')
        if web3_provider is None:
            raise ValueError('TEST_WEB3_PROVIDER_URL must be set')

        # Create client that will be used to fund the new user.
        source_client = Client(
            host='',
            eth_private_key=source_private_key,
            web3_provider=web3_provider,
        )

        # Create an Ethereum account and STARK keys for the new user.
        web3_account = Web3(None).eth.account.create()
        ethereum_address = web3_account.address
        eth_private_key = web3_account.key
        stark_private_key = generate_private_key_hex_unsafe()

        # Fund the new user with ETH and USDC.
        fund_eth_hash = source_client.eth.transfer_eth(
            to_address=ethereum_address,
            human_amount=0.001,
        )
        fund_usdc_hash = source_client.eth.transfer_token(
            to_address=ethereum_address,
            human_amount=2,
        )
        print('Waiting for funds...')
        source_client.eth.wait_for_tx(fund_eth_hash)
        source_client.eth.wait_for_tx(fund_usdc_hash)
        print('...done.')

        # Create client for the new user.
        client = Client(
            host=HOST,
            network_id=NETWORK_ID,
            stark_private_key=stark_private_key,
            eth_private_key=eth_private_key,
            web3_provider=web3_provider,
        )

        # Onboard the user.
        res = client.onboarding.create_user()
        api_key_credentials = res['apiKey']

        print('eth_private_key', eth_private_key)
        print('stark_private_key', stark_private_key)
        print('client.api_key_credentials', client.api_key_credentials)

        # Get the user.
        get_user_result = client.private.get_user()
        assert get_user_result['user'] == {
            'ethereumAddress': ethereum_address.lower(),
            'isRegistered': False,
            'email': None,
            'username': None,
            'userData': {},
            'makerFeeRate': '0.0005',
            'takerFeeRate': '0.0015',
            'makerVolume30D': '0',
            'takerVolume30D': '0',
            'fees30D': '0',
        }

        # Get the registration signature.
        registration_result = client.private.get_registration()
        signature = registration_result['signature']
        assert re.match('0x[0-9a-f]{130}$', signature) is not None, (
            'Invalid registration result: {}'.format(registration_result)
        )

        # Register the user on-chain.
        registration_tx_hash = client.eth.register_user(signature)
        print('Waiting for registration...')
        client.eth.wait_for_tx(registration_tx_hash)
        print('...done.')

        # Set the user's username.
        username = 'integration_user_{}'.format(int(time.time()))
        client.private.update_user(username=username)

        # Create a second account under the same user.
        #
        # NOTE: Support for multiple accounts under the same user is limited.
        # The frontend does not currently support mutiple accounts per user.
        stark_private_key_2 = generate_private_key_hex_unsafe()
        stark_public_key_2, stark_public_key_y_coordinate_2 = (
            private_key_to_public_key_pair_hex(stark_private_key_2)
        )

        client.private.create_account(
            stark_public_key=stark_public_key_2,
            stark_public_key_y_coordinate=stark_public_key_y_coordinate_2,
        )

        # Get the primary account.
        get_account_result = client.private.get_account(
            ethereum_address=ethereum_address,
        )
        account = get_account_result['account']
        assert int(account['starkKey'], 16) == int(client.stark_public_key, 16)

        # Get all accounts.
        get_all_accounts_result = client.private.get_accounts()
        get_all_accounts_public_keys = [
            a['starkKey'] for a in get_all_accounts_result['accounts']
        ]
        assert int(client.stark_public_key, 16) in [
            int(k, 16) for k in get_all_accounts_public_keys
        ]

        # TODO: Fix.
        # assert int(stark_public_key_2, 16) in [
        #     int(k, 16) for k in get_all_accounts_public_keys
        # ]

        # Get positions.
        get_positions_result = client.private.get_positions(market='BTC-USD')
        assert get_positions_result == {'positions': []}

        # Set allowance on the Starkware perpetual contract, for the deposit.
        approve_tx_hash = client.eth.set_token_max_allowance(
            client.eth.get_exchange_contract().address,
        )
        print('Waiting for allowance...')
        client.eth.wait_for_tx(approve_tx_hash)
        print('...done.')

        # Send an on-chain deposit.
        deposit_tx_hash = client.eth.deposit_to_exchange(
            account['positionId'],
            3,
        )
        print('Waiting for deposit...')
        client.eth.wait_for_tx(deposit_tx_hash)
        print('...done.')

        # Wait for the deposit to be processed.
        print('Waiting for deposit to be processed on dYdX...')
        wait_for_condition(
            lambda: len(client.private.get_transfers()['transfers']) > 0,
            True,
            60,
        )
        print('...transfer was recorded, waiting for confirmation...')
        wait_for_condition(
            lambda: client.private.get_account()['account']['quoteBalance'],
            '2',
            180,
        )
        print('...done.')

        # Post an order.
        one_minute_from_now_iso = epoch_seconds_to_iso(time.time() + 60)
        create_order_result = client.private.create_order(
            position_id=account['positionId'],
            market=constants.MARKET_BTC_USD,
            side=constants.ORDER_SIDE_BUY,
            order_type=constants.ORDER_TYPE_LIMIT,
            post_only=False,
            size='10',
            price='1000',
            limit_fee='0.1',
            expiration=one_minute_from_now_iso,
        )

        # Get the order.
        order_id = create_order_result['order']['id']
        get_order_result = client.private.get_order_by_id(order_id)
        assert get_order_result['order']['market'] == constants.MARKET_BTC_USD

        # Cancel the order.
        client.private.cancel_order(order_id)

        # Cancel all orders.
        client.private.cancel_all_orders()

        # Get open orders.
        get_orders_result = client.private.get_orders(
            market=constants.MARKET_BTC_USD,
            status=constants.POSITION_STATUS_OPEN,
        )
        assert get_orders_result == {'orders': []}

        # Get fills.
        client.private.get_fills(
            market=constants.MARKET_BTC_USD,
        )

        # Initiate a regular (slow) withdrawal.
        expiration_epoch_seconds = time.time() + SEVEN_DAYS_S + 60
        client.private.create_withdrawal(
            position_id=account['positionId'],
            amount='1',
            asset=constants.ASSET_USDC,
            to_address=ethereum_address,
            expiration_epoch_seconds=expiration_epoch_seconds,
        )

        # Get deposits.
        deposits_result = client.private.get_transfers(
            transfer_type=constants.ACCOUNT_ACTION_DEPOSIT,
        )
        assert len(deposits_result['transfers']) == 1

        # Get withdrawals.
        withdrawals_result = client.private.get_transfers(
            transfer_type=constants.ACCOUNT_ACTION_WITHDRAWAL,
        )
        assert len(withdrawals_result['transfers']) == 1

        # Get funding payments.
        client.private.get_funding_payments(
            market=constants.MARKET_BTC_USD,
        )

        # Register a new API key.
        create_api_key_result = client.eth_private.create_api_key()
        new_api_key_credentials = create_api_key_result['apiKey']

        # Get all API keys.
        get_api_keys_result = client.private.get_api_keys()
        api_keys_public_keys = [
            a['key'] for a in get_api_keys_result['apiKeys']
        ]
        assert api_key_credentials['key'] in api_keys_public_keys

        # Delete an API key.
        client.eth_private.delete_api_key(
            api_key=new_api_key_credentials['key'],
            ethereum_address=ethereum_address,
        )

        # Get all API keys after the deletion.
        get_api_keys_result_after = client.private.get_api_keys()
        assert len(get_api_keys_result_after['apiKeys']) == 1

        # Initiate a fast withdrawal of USDC.
        expiration_epoch_seconds = time.time() + SEVEN_DAYS_S + 60
        client.private.create_fast_withdrawal(
            position_id=account['positionId'],
            credit_asset='USDC',
            credit_amount='1',
            debit_amount='2',
            to_address=ethereum_address,
            lp_position_id=LP_POSITION_ID,
            lp_stark_public_key=LP_PUBLIC_KEY,
            client_id='mock-client-id',
            expiration_epoch_seconds=expiration_epoch_seconds,
        )

        # Send an on-chain withdraw.
        withdraw_tx_hash = client.eth.withdraw()
        print('Waiting for withdraw...')
        client.eth.wait_for_tx(withdraw_tx_hash)
        print('...done.')

        # Wait for the withdraw to be processed.
        print('Waiting for withdraw to be processed on dYdX...')
        wait_for_condition(
            lambda: len(client.private.get_transfers()['transfers']) > 0,
            True,
            60,
        )
        print('...transfer was recorded, waiting for confirmation...')
        wait_for_condition(
            lambda: client.private.get_account()['account']['quoteBalance'],
            '2',
            180,
        )
        print('...done.')

        # Send an on-chain withdraw_to.
        withdraw_to_tx_hash = client.eth.withdraw_to(
            recipient=ethereum_address,
        )
        print('Waiting for withdraw_to...')
        client.eth.wait_for_tx(withdraw_to_tx_hash)
        print('...done.')

        # Wait for the withdraw_to to be processed.
        print('Waiting for withdraw_to to be processed on dYdX...')
        wait_for_condition(
            lambda: len(client.private.get_transfers()['transfers']) > 0,
            True,
            60,
        )
        print('...transfer was recorded, waiting for confirmation...')
        wait_for_condition(
            lambda: client.private.get_account()['account']['quoteBalance'],
            '2',
            180,
        )
        print('...done.')



"""Constants related to creating hashes of Starkware objects."""

ONE_HOUR_IN_SECONDS = 60 * 60
ORDER_SIGNATURE_EXPIRATION_BUFFER_HOURS = 24 * 7  # Seven days.

TRANSFER_PREFIX = 4
TRANSFER_PADDING_BITS = 81
CONDITIONAL_TRANSFER_PADDING_BITS = 81
CONDITIONAL_TRANSFER_PREFIX = 5
ORDER_PREFIX = 3
ORDER_PADDING_BITS = 17
WITHDRAWAL_PADDING_BITS = 49
WITHDRAWAL_PREFIX = 6

# Note: Fees are not supported for conditional transfers or transfers.
TRANSFER_FEE_ASSET_ID = 0
TRANSFER_MAX_AMOUNT_FEE = 0

CONDITIONAL_TRANSFER_FEE_ASSET_ID = 0
CONDITIONAL_TRANSFER_MAX_AMOUNT_FEE = 0

TRANSFER_FIELD_BIT_LENGTHS = {
    "asset_id": 250,
    "receiver_public_key": 251,
    "position_id": 64,
    "quantums_amount": 64,
    "nonce": 32,
    "expiration_epoch_hours": 32,
}

CONDITIONAL_TRANSFER_FIELD_BIT_LENGTHS = {
    "asset_id": 250,
    "receiver_public_key": 251,
    "position_id": 64,
    "condition": 251,
    "quantums_amount": 64,
    "nonce": 32,
    "expiration_epoch_hours": 32,
}

ORDER_FIELD_BIT_LENGTHS = {
    "asset_id_synthetic": 128,
    "asset_id_collateral": 250,
    "asset_id_fee": 250,
    "quantums_amount": 64,
    "nonce": 32,
    "position_id": 64,
    "expiration_epoch_hours": 32,
}

WITHDRAWAL_FIELD_BIT_LENGTHS = {
    "asset_id": 250,
    "position_id": 64,
    "nonce": 32,
    "quantums_amount": 64,
    "expiration_epoch_hours": 32,
}

ORACLE_PRICE_FIELD_BIT_LENGTHS = {
    "asset_name": 128,
    "oracle_name": 40,
    "price": 120,
    "timestamp_epoch_seconds": 32,
}



from web3 import Web3

from dydx3.constants import NETWORK_ID_MAINNET
from dydx3.eth_signing import SignWithWeb3
from dydx3.eth_signing import SignWithKey
from dydx3.eth_signing import SignEthPrivateAction

GANACHE_ADDRESS = '0x90F8bf6A479f320ead074411a4B0e7944Ea8c9C1'
GANACHE_PRIVATE_KEY = (
    '0x4f3edf983ac636a65a842ce7c78d9aa706d3b113bce9c46f30d7d21715b23b1d'
)
PARAMS = {
    'method': 'POST',
    'request_path': 'v3/test',
    'body': '{}',
    'timestamp': '2021-01-08T10:06:12.500Z',
}

EXPECTED_SIGNATURE = (
    '0x3ec5317783b313b0acac1f13a23eaaa2fca1f45c2f395081e9bfc20b4cc1acb17e'
    '3d755764f37bf13fa62565c9cb50475e0a987ab0afa74efde0b3926bb7ab9d1b00'
)


class TestApiKeyAction():

    def test_sign_via_local_node(self):
        web3 = Web3()  # Connect to a local Ethereum node.
        signer = SignWithWeb3(web3)

        action_signer = SignEthPrivateAction(signer, NETWORK_ID_MAINNET)
        signature = action_signer.sign(GANACHE_ADDRESS, **PARAMS)
        assert action_signer.verify(
            signature,
            GANACHE_ADDRESS,
            **PARAMS,
        )
        assert signature == EXPECTED_SIGNATURE

    def test_sign_via_account(self):
        web3 = Web3(None)
        web3_account = web3.eth.account.create()
        signer = SignWithKey(web3_account.key)

        action_signer = SignEthPrivateAction(signer, NETWORK_ID_MAINNET)
        signature = action_signer.sign(signer.address, **PARAMS)
        assert action_signer.verify(
            signature,
            signer.address,
            **PARAMS,
        )

    def test_sign_via_private_key(self):
        signer = SignWithKey(GANACHE_PRIVATE_KEY)

        action_signer = SignEthPrivateAction(signer, NETWORK_ID_MAINNET)
        signature = action_signer.sign(GANACHE_ADDRESS, **PARAMS)
        assert action_signer.verify(
            signature,
            GANACHE_ADDRESS,
            **PARAMS,
        )
        assert signature == EXPECTED_SIGNATURE
